"""
Functions for finding pairs within a certain search distance "err".

Very fast method based on hashing.
"""
from __future__ import print_function, division
from nwaylib.fastskymatch import dist, array2fits, match_multiple, wraptable2fits, fits_from_columns
import astropy.io.fits as pyfits
import numpy
import nwaylib.progress as progress
import nwaylib.logger as logger
import astropy.units as u
from astropy.coordinates import SkyCoord, SkyOffsetFrame
import numpy.testing as test

def test_dist():
	
	d = dist((53.15964508, -27.92927742), (53.15953445, -27.9313736))
	print('distance', d)
	assert not numpy.isnan(d)
	
	ra = numpy.array([ 53.14784241,  53.14784241,  53.14749908, 53.16559982,     53.19423676,  53.1336441 ])
	dec = numpy.array([-27.79363823, -27.79363823, -27.81790352, -27.79622459,      -27.70860672, -27.76327515])
	ra2 = numpy.array([ 53.14907837,  53.14907837,  53.1498642 , 53.16150284,     53.19681549,  53.13626862])
	dec2 = numpy.array([-27.79297447, -27.79297447, -27.81404877, -27.79223251,  -27.71365929, -27.76314735])

	d = dist((ra, dec), (ra2, dec2))
	print('distance', d)
	assert not numpy.isnan(d).any()

def run_match(nfiles, ngen=40):
	numpy.random.seed(0)

	def gen():
		ra  = numpy.random.uniform(size=ngen)
		dec = numpy.random.uniform(size=ngen)
		return numpy.array(list(zip(ra, dec)), dtype=[('ra', 'f'), ('dec', 'f')])
	
	filenames = ['test_input_%d.fits' % i for i in range(nfiles)]

	for fitsname in filenames:
		print('generating toy data for %s!' % fitsname)
		hdulist = array2fits(gen(), fitsname.replace('.fits', ''))
		hdulist[0].header['GENERAT'] = 'match test table, random'
		hdulist.writeto(fitsname, **progress.kwargs_overwrite_true)

	tables = [pyfits.open(fitsname)[1] for fitsname in filenames]
	table_names = [t.name for t in tables]
	tables = [t.data for t in tables]

	err = 0.03
	fits_formats = []
	for table in tables:
		fits_formats.append([c.format for c in table.columns])

	results, columns, header = match_multiple(tables, table_names, err, fits_formats, logger=logger.NormalLogger())
	tbhdu = fits_from_columns(pyfits.ColDefs(columns))

	hdulist = wraptable2fits(tbhdu, 'MATCH')
	hdulist[0].header['ANALYSIS'] = 'match table from' + ', '.join(table_names)
	hdulist[0].header['INPUT'] = ', '.join(filenames)
	hdulist.writeto('test_match%d.fits' % nfiles, **progress.kwargs_overwrite_true)

	print('plotting')
	tbhdu = pyfits.open('test_match%d.fits' % nfiles)[1]
	for table_name in table_names:
		ra = tbhdu.data['%s_RA'  % table_name]
		dec = tbhdu.data['%s_DEC'  % table_name]
		
		print(ra.shape, dec.shape)
		assert len(ra)>20
		assert len(dec)==len(ra)

def test_dist_astropyconsistent():
	# positions:
	dec = numpy.linspace(0, 90, 4000)[1:-1]
	ra = 0
	shift = 1e-5
	offset_ra  = shift
	offset_dec = 0 * shift

	# precision:
	sigma_ra  = 1 * u.arcsec
	sigma_dec = 1 * u.arcsec
	sigma_ra2  = sigma_ra / 10
	sigma_dec2 = sigma_dec / 10
	sigma_ra_tot = (sigma_ra**-2 + sigma_ra2**-2)**-0.5
	sigma_dec_tot = (sigma_dec**-2 + sigma_dec2**-2)**-0.5

	a = SkyCoord(ra, dec, frame="icrs", unit="deg")
	b = SkyCoord(ra+offset_ra, dec+offset_dec, frame="icrs", unit="deg")
	localframe = SkyOffsetFrame(origin=a)
	na = a.transform_to(localframe)
	nb = b.transform_to(localframe)
	
	# compute projected distance on the sky using our function
	d = dist((ra, dec), (ra+offset_ra, dec+offset_dec))
	# compute comparison with astropy
	diff = b.separation(a).to(u.deg)

	# compute simple flat version
	dra  = na.lon - nb.lon
	ddec = na.lat - nb.lat
	simplelocaldiff = (dra**2 + ddec**2)**0.5
	test.assert_almost_equal(diff / u.deg, simplelocaldiff / u.deg)

	

def test_match_2():
	run_match(2, ngen=1000)

def test_match_3():
	run_match(3, ngen=400)

def test_match_4():
	run_match(4, ngen=100)

def test_match_5():
	run_match(5, ngen=40)

import nwaylib
import nwaylib.logger as logger

def create_affine_invariance_test(ngen=1000, dilution_factor=0, factor=1):
	numpy.random.seed(1)
	print()
	print("create_affine_invariance_test(ngen=", ngen, ", dilution_factor=",dilution_factor, ", factor=", factor)
	nfiles = 2
	lo = 0.0
	hi = 0.1 * factor
	errs = [5 * factor, 0.001]
	area = (hi - lo)**2
	IDs = numpy.arange(ngen)
	ra_true = numpy.random.uniform(size=ngen) * hi
	dec_true = numpy.random.uniform(size=ngen) * hi

	def genwithextra(err, nextra):
		ra  = numpy.random.normal(0, 1, size=len(ra_true)) * err / 3600 + ra_true
		dec = numpy.random.normal(0, 1, size=len(dec_true)) * err / 3600 + dec_true
		print('pos:', ra[:4], dec[:4], '+-', err)
		if nextra > 0:
			ID_extra = numpy.arange(nextra) + 1000000
			ra_extra = numpy.random.uniform(lo, hi, size=nextra)
			dec_extra = numpy.random.uniform(lo, hi, size=nextra)
		else:
			ID_extra, ra_extra, dec_extra = [], [], []
		return numpy.array(list(zip(IDs, ra, dec, err * numpy.ones(ngen))) + list(zip(ID_extra, ra_extra, dec_extra, err * numpy.ones(nextra))), 
			dtype=[('ID', 'i'), ('ra', 'f'), ('dec', 'f'), ('err', 'f')])

	filenames = ['test_input_%d_%d_%d.fits' % (factor, dilution_factor, i) for i in range(nfiles)]
	
	nway_inputs = []

	for fitsname, err, f, table_name in zip(filenames, errs, [0, dilution_factor], ["X", "O"]):
		print('generating toy data for %s!' % fitsname)
		table_data = genwithextra(err=err, nextra=ngen * f)
		hdulist = array2fits(table_data, table_name)
		hdulist[0].header['GENERAT'] = 'match test table, random'
		hdulist[1].header['SKYAREA'] = area
		hdulist.writeto(fitsname, **progress.kwargs_overwrite_true)
		nway_inputs.append(dict(name=table_name, ra=table_data['ra'], dec=table_data['dec'], error=table_data['err'], area=area, mags=[], maghists=[], magnames=[]))
	
	result = nwaylib.nway_match(nway_inputs,
		match_radius = errs[0] * 10, # in arcsec
		prior_completeness = 1.0,
		logger=logger.NullOutputLogger(),
	)

	p_any = result['prob_has_match'][result['match_flag'] == 1]
	print(numpy.median(p_any), numpy.min(p_any), numpy.max(p_any),)


def run_affine_invariance_test(ngen=10000):
	# check that a 10x field side length increase with same error increase
	# leads to the same p_any distribution
	
	for dilution_factor in 0,: #1, 10:
		print("="*60)
		print("Dilution:", dilution_factor)
		for factor in 1, 10, 100:
			create_affine_invariance_test(ngen=ngen, dilution_factor=dilution_factor, factor=factor)

			

if __name__ == '__main__':
	run_affine_invariance_test()



import tqdm

def bar(**kwargs):
	return tqdm.tqdm
	
	
import inspect
import astropy.io.fits as pyfits
args = inspect.signature(pyfits.writeto).parameters if hasattr(inspect, 'signature') else inspect.getargspec(pyfits.writeto).args
if 'overwrite' in args:
	arg_overwrite = 'overwrite'
else:
	arg_overwrite = 'clobber'
kwargs_overwrite_true = {arg_overwrite:True}
kwargs_overwrite_false = {arg_overwrite:False}




"""Multiway association between astrometric catalogues"""
from __future__ import print_function, division

import numpy
from numpy import log10, pi
import pandas
from collections import OrderedDict

from .logger import NullOutputLogger, NormalLogger
from . import fastskymatch as match
from . import bayesdistance as bayesdist
from . import magnitudeweights as magnitudeweights

__author__ = """Johannes Buchner"""
__email__ = 'johannes.buchner.acad@gmx.com'
__version__ = '4.5.4'


class UndersampledException(Exception):
	pass
class EmptyResultException(Exception):
	pass

def nway_match(match_tables, match_radius, prior_completeness,
	mag_include_radius=None, mag_exclude_radius=None, magauto_post_single_minvalue=0.9,
	prob_ratio_secondary = 0.5,
	min_prob=0., consider_unrelated_associations=True, 
	store_mag_hists=True,
	logger=NormalLogger()):
	"""
	match_tables: list of catalogues, each a dict with entries:
		- name (short catalog name, no spaces, used in output columns)
		- ra (RA in degrees)
		- dec (dec in degrees)
		- error (positional error in arcsec)
		- area (sky area covered by the catalogue in square degrees)
		- mags (list of additional columns to consider as priors)
		- magnames (short name for each entry in mags, used in output columns)
		- maghists (list of prior information for each entry in mags)
		  use None to automatically build a histogram of target/non-target sources.
		  Otherwise, supply the histogram manually: bins_lo, bins_hi, hist_sel, hist_all.
	
	match_radius: maximum radius in arcsec to consider.
		More distant counterparts are cut off. 
		Set to a very large value, e.g. 5 times the largest positional error.
		Setting to larger values does not change the result, but
		setting to smaller values improves performance.
	
	prior_completeness: expected fraction of sources in the primary catalogue 
		that are expected to have a counterpart (e.g., 90%).
		If an array is passed, completeness for each catalog. First
		entry has to be 1.
	
	mag_include_radius:
		search radius for building the magnitude histogram of target sources. If None (default), the Bayesian posterior is used
	
	mag_exclude_radius:
		exclusion radius for building the magnitude histogram of field sources. If None (default), mag_include_radius is used
	
	magauto_post_single_minvalue:
		minimum posterior probability (default: 0.9) for the magnitude histogram of secure target sources. Used in the Bayesian procedure.
	
	prob_ratio_secondary: for each primary catalogue entry, the most probable associations is marked with match_flag=1. Comparably good solutions receive match_flag=2 (all others are match_flag=0). The prob_ratio_secondary parameter controls the lowest probability ratio p(secondary)/p(primary) still considered comparable.
	
	min_prob: Minimum probability for truncating the output table.
		default: 0 (no truncation)
	
	consider_unrelated_associations: Correction for missing sources in 
		partial assocations. Keep to True.
	
	store_mag_hists: Write constructed mag hists to file (filename based on table name and mags).
	
	logger: NormalLogger for stderr output and progress bars, NullOutputLogger if silent
	"""
	if mag_exclude_radius is None:
		mag_exclude_radius = mag_include_radius
	if mag_include_radius is not None:
		if mag_include_radius >= match_radius:
			logger.warn('WARNING: magnitude radius is very large (>= matching radius). Consider using a smaller value.')

	ncats = len(match_tables)
	
	table, resultstable, separations, errors = _create_match_table(match_tables, match_radius, logger=logger)

	if not len(table) > 0:
		raise EmptyResultException('No matches.')
	
	source_densities, source_densities_plus = _compute_source_densities(match_tables, logger=logger)

	# first pass: find secure matches and secure non-matches

	prior, log_bf = _compute_single_log_bf(match_tables, source_densities, source_densities_plus, table, separations, errors, prior_completeness, logger=logger)
	table = table.assign(
		dist_bayesfactor_uncorrected=log_bf,
		dist_bayesfactor=log_bf
	)
	
	if consider_unrelated_associations:
		_correct_unrelated_associations(table, separations, errors, ncats, source_densities, source_densities_plus, logger=logger)
		log_bf = table['dist_bayesfactor']

	# add the additional columns
	post = bayesdist.posterior(prior, log_bf)
	table = table.assign(dist_post=post)

	# find magnitude biasing functions
	table, total = _apply_magnitude_biasing(match_tables, table, mag_include_radius, mag_exclude_radius, magauto_post_single_minvalue, store_mag_hists, logger=logger)

	table = _compute_final_probabilities(match_tables, table, prob_ratio_secondary, prior, total, logger=logger)
	
	table = _truncate_table(table, min_prob, logger=logger)
	
	return table


def _create_match_table(match_tables, match_radius, logger):
	# first match input catalogues, compute possible combinations in match_radius
	ratables = [(t['ra'], t['dec']) for t in match_tables]
	table_names = [t['name'] for t in match_tables]

	resultstable = match.crossproduct(ratables, match_radius / 60. / 60, logger=logger)
	#results = resultstable.view(dtype=[(t['name'], resultstable.dtype) for t in match_tables]).reshape((-1,))
	nresults = len(resultstable)
	keys = []
	columns = []
	for i, t in enumerate(match_tables):
		keys.append(t['name'])
		columns.append(resultstable[:,i])

	# matrix of separations
	separations = []
	errors = []
	invalid_separations = numpy.ones(nresults) * numpy.nan
	logger.log('    adding angular separation columns')
	max_separation = numpy.zeros(nresults)
	for i in range(len(match_tables)):
		a_ra  = ratables[i][0][resultstable[:,i]]
		a_dec = ratables[i][1][resultstable[:,i]]
		errors.append(match_tables[i]['error'][resultstable[:,i]])
		row = []
		for j in range(len(match_tables)):
			if i < j:
				b_ra  = ratables[j][0][resultstable[:,j]]
				b_dec = ratables[j][1][resultstable[:,j]]
				col = match.dist((a_ra, a_dec), (b_ra, b_dec))
				assert not numpy.isnan(col).any(), ['%d distances are nan' % numpy.isnan(col).sum(), 
					a_ra[numpy.isnan(col)], a_dec[numpy.isnan(col)], 
					b_ra[numpy.isnan(col)], b_dec[numpy.isnan(col)]]
				
				assert not (a_ra == b_ra).all()
				# store distance in arcsec
				col[resultstable[:,i] == -1] = numpy.nan
				col[resultstable[:,j] == -1] = numpy.nan
				col[a_ra == -99] = numpy.nan
				col[b_ra == -99] = numpy.nan
				col_arcsec = col * 60 * 60
				keys.append("Separation_%s_%s" % (table_names[i], table_names[j]))
				columns.append(col_arcsec)
				max_separation = numpy.nanmax([col_arcsec, max_separation], axis=0)
				row.append(col_arcsec)
			else:
				# mark the upper right triangle of the matrix as invalid data
				row.append(invalid_separations)
		separations.append(row)
		del row

	keys.append('Separation_max')
	columns.append(max_separation)
	keys.append('ncat')
	columns.append((resultstable > -1).sum(axis=1))

	# now truncate all columns:
	mask = max_separation < match_radius
	columns = [c[mask] for c in columns]
	errors = [e[mask] for e in errors]
	separations = [[cell[mask] for cell in row] for row in separations]
	#results = results[mask]
	resultstable = resultstable[mask,:]
	nresults = len(resultstable)

	logger.log('matching: %6d matches after filtering by search radius' % mask.sum())

	# now we have columns, which contains the distance information.
	assert len(columns) == len(keys), (len(columns), len(keys),  keys)

	table = pandas.DataFrame(OrderedDict(zip(keys, columns)))
	assert len(table) == nresults, (len(table), nresults,  mask.sum())

	return table, resultstable, separations, errors

def _compute_source_densities(match_tables, logger):
	source_densities = []
	source_densities_plus = []
	for i, match_table in enumerate(match_tables):
		n = len(match_table['ra'])
		area = match_table['area'] * 1.0 # in square degrees
		area_total = (4 * pi * (180 / pi)**2)
		density = n / area * area_total
		logger.log('%s "%s" (%d), density gives %.2e objects on entire sky' % ('Primary catalogue' if i == 0 else 'Catalogue', match_table['name'], n, density))
		# this takes into account that the source may be absent
		density_plus = (n + 1) / area * area_total
		source_densities.append(density)
		source_densities_plus.append(density_plus)

	# source can not be absent in primary catalogue
	source_densities_plus[0] = source_densities[0]
	source_densities_plus = numpy.array(source_densities_plus)
	source_densities = numpy.array(source_densities)
	return source_densities, source_densities_plus

def _compute_single_log_bf(match_tables, source_densities, source_densities_plus, table, separations, errors, prior_completeness, logger):
	logger.log('Computing distance-based probabilities ...')
	ncats = len(match_tables)

	if numpy.shape(prior_completeness) == ():
		prior_completeness = numpy.array([1.0] + [float(prior_completeness)**(1./(ncats-1)) for i in range(1, ncats)])
	
	if len(prior_completeness) != ncats:
		raise Exception('Prior completeness needs one value per catalog. Received "%s".' % prior_completeness)
	assert prior_completeness[0] == 1.0

	log_bf = numpy.zeros(len(table)) * numpy.nan
	prior = numpy.zeros(len(table)) * numpy.nan
	# handle all cases (also those with missing counterparts in some catalogues)
	for case in range(2**(ncats-1)):
		table_mask = numpy.array([True] + [(case // 2**(ti)) % 2 == 0 for ti in range(len(match_tables)-1)])
		# select those cases
		mask = True
		for i in range(1, len(match_tables)):
			if table_mask[i]: # require not nan
				mask = numpy.logical_and(mask, ~numpy.isnan(separations[0][i]))
			else:
				mask = numpy.logical_and(mask, numpy.isnan(separations[0][i]))
		# select errors
		errors_selected = [e[mask] for e, m in zip(errors, table_mask) if m]
		separations_selected = [[cell[mask] for cell, m in zip(row, table_mask) if m] 
			for row, m2 in zip(separations, table_mask) if m2]
		# here we should call the elliptical error variant if errors is a 2d array
		r = bayesdist.log_bf(separations_selected, errors_selected)
		assert r.shape == separations_selected[0][0].shape, (r.shape, separations_selected[0][0].shape)
		assert r.shape == errors_selected[0].shape, (r.shape, errors_selected[0].shape)
		assert r.shape == mask.sum(), (r.shape, mask.sum(), mask.shape)
		log_bf[mask] = r

		prior[mask] = source_densities[0] * numpy.product(prior_completeness[table_mask]) / numpy.product(source_densities_plus[table_mask])
		assert numpy.isfinite(prior[mask]).all(), (source_densities, prior_completeness[table_mask], numpy.product(source_densities_plus[table_mask]))


	assert numpy.isfinite(prior).all(), (prior, log_bf)
	assert numpy.isfinite(log_bf).all(), (prior, log_bf)
	return prior, log_bf

def _correct_unrelated_associations(table, separations, errors, ncats, source_densities, source_densities_plus, logger):
	logger.log('    correcting for unrelated associations ...')
	# correct for unrelated associations
	# identify those in need of correction
	# two unconsidered catalogues are needed for an unrelated association
	primary_id = table[table.columns[0]]
	#candidates = numpy.unique(primary_id[(ncat <= ncats - 2).values])
	pbar = logger.progress()
	for primary_id_key, group in pbar(table.groupby(primary_id, sort=False)):
		ncat_here = group['ncat']
		# list which ones we are missing
		best_logpost = 0
		# go through more complex associations
		for j in group[ncat_here > 2].index.values:
			missing_cats = [k for k, sep in enumerate(separations[0]) if numpy.isnan(sep[j])]
			# check if this association has sufficient overlap with the one we are looking for
			# it must contain at least two of the catalogues we are missing
			augmented_cats = []
			for k in missing_cats:
				if not numpy.isnan(separations[0][k][j]):
					augmented_cats.append(k)
			n_augmented_cats = len(augmented_cats)
			if n_augmented_cats >= 2:
				# ok, this is helpful.
				# identify the separations and errors
				separations_selected = [[[separations[k][k2][j]] for k2 in augmented_cats] for k in augmented_cats]
				errors_selected = [[errors[k][j]] for k in augmented_cats]
				# identify the prior
				prior_j = source_densities[augmented_cats[0]] / numpy.product(source_densities_plus[augmented_cats])
				# compute a log_bf
				log_bf_j = bayesdist.log_bf(numpy.array(separations_selected), numpy.array(errors_selected))
				logpost_j = bayesdist.unnormalised_log_posterior(prior_j, log_bf_j, n_augmented_cats)
				if logpost_j > best_logpost:
					#print('post:', logpost_j, log_bf_j, prior_j)
					best_logpost = logpost_j

		# ok, we have our correction factor, best_logpost
		# lets multiply it onto log_bf
		if best_logpost > 0:
			group['dist_bayesfactor'] += best_logpost

def _apply_magnitude_biasing(match_tables, table, mag_include_radius, mag_exclude_radius, magauto_post_single_minvalue, store_mag_hists, logger):
	biases = {}
	for i, t in enumerate(match_tables):
		table_name = t['name']
		for magvals, maghist, magname in zip(t['mags'], t['maghists'], t['magnames']):
			col_name = magname
			col = "%s_%s" % (table_name, col_name)
			mag = "%s:%s" % (table_name, col_name)
			logger.log('Incorporating bias "%s" ...' % mag)
			
			res = table[table.columns[i]].values
			res_defined = res != -1
			# get magnitudes of all
			# mark -99 as undefined
			mag_all = magvals
			mag_all[mag_all == -99] = numpy.nan
		
			# get magnitudes of selected
			mask_all = numpy.isfinite(mag_all)
			
		
			if maghist is None:
				if mag_include_radius is not None:
					selection = table['Separation_max'].values < mag_include_radius
					selection_possible = table['Separation_max'].values < mag_exclude_radius
					selection_weights = numpy.ones(len(selection))
				else:
					selection = (table['dist_post'] > magauto_post_single_minvalue).values
					selection_weights = table['dist_post'].values
					selection_possible = (table['dist_post'] > 0.01).values
				
				# ignore cases where counterpart is missing
				assert res_defined.shape == selection.shape, (res_defined.shape, selection.shape)
				selection = numpy.logical_and(selection, res_defined)
				selection_weights = selection_weights[res_defined]
				selection_possible = numpy.logical_and(selection_possible, res_defined)
				
				#print '   selection', selection.sum(), selection_possible.sum(), (-selection_possible).sum()
				
				#rows = results[table_name][selection].tolist()
				assert selection.shape == res.shape, (selection.shape, res.shape)
				rows, unique_indices = numpy.unique(res[selection], return_index=True)
				rows_weights = selection_weights[unique_indices]
				
				assert len(rows) > 0, 'No magnitude values within radius for "%s".' % mag
				mag_sel = magvals[rows]
				mag_sel_weights = rows_weights
				
				# remove vaguely possible options from alternative histogram
				assert selection_possible.shape == res.shape, (selection_possible.shape, res.shape)
				#print(res, selection, selection_possible)
				rows_possible = numpy.unique(res[selection_possible])
				mask_others = mask_all.copy()
				mask_others[rows_possible] = False
				
				# all options in the total (field+target sources) histogram
				mask_sel = ~numpy.logical_or(numpy.isnan(mag_sel), numpy.isinf(mag_sel))

				#print '      non-nans: ', mask_sel.sum(), mask_others.sum()

				logger.log('magnitude histogram of column "%s": %d secure matches, %d insecure matches and %d secure non-matches of %d total entries (%d valid)' % (col, mask_sel.sum(), len(rows_possible), mask_others.sum(), len(mag_all), mask_all.sum()))
				
				# make function fitting to ratio shape
				bins, hist_sel, hist_all = magnitudeweights.adaptive_histograms(mag_all[mask_others], mag_sel[mask_sel], weights=mag_sel_weights[mask_sel])
				if store_mag_hists:
					logger.log('magnitude histogram stored to "%s".' % (mag.replace(':', '_') + '_fit.txt'))
					with open(mag.replace(':', '_') + '_fit.txt', 'wb') as f:
						f.write(b'# lo hi selected others\n')
						numpy.savetxt(f,
							numpy.transpose([bins[:-1], bins[1:], hist_sel, hist_all]), 
							fmt = ["%10.5f"]*4)
				if mask_sel.sum() < 100:
					raise UndersampledException('ERROR: too few secure matches (%d) to make a good histogram. If you are sure you want to use this poorly sampled histogram, replace "auto" with the filename. You can also decrease the mag-auto-minprob parameter.' % mask_sel.sum())
			else:
				logger.log('magnitude histogramming: using user-supplied histogram for "%s"' % (col))
				bins_lo, bins_hi, hist_sel, hist_all = maghist #numpy.loadtxt(magfile).transpose()
				bins = numpy.array(list(bins_lo) + [bins_hi[-1]])
			func = magnitudeweights.fitfunc_histogram(bins, hist_sel, hist_all)
			if store_mag_hists:
				magnitudeweights.plot_fit(bins, hist_sel, hist_all, func, mag)
			magcol = magvals[res]
			magcol[~numpy.logical_and(res_defined, numpy.isfinite(magcol))] = -99
			#magcol[~res_defined] = -99
			weights = log10(func(magcol))
			# undefined magnitudes do not contribute
			weights[numpy.isnan(weights)] = 0
			biases[col] = weights

	# add the bias columns
	table = table.assign(**{'bias_%s' % col:10**weights for col, weights in biases.items()})
	log_bf = table['dist_bayesfactor'].values
	total = log_bf + sum(biases.values())
	
	return table, total

def _compute_final_probabilities(match_tables, table, prob_ratio_secondary, prior, total, logger):
	logger.log('')
	logger.log('Computing final probabilities ...')

	# add the posterior column
	post = bayesdist.posterior(prior, total)
	table = table.assign(p_single=post)

	# compute weights for group posteriors
	# 4pi comes from Eq. 
	ncat = table['ncat'].values
	log_post_weight = bayesdist.unnormalised_log_posterior(prior, total, ncat)
	table = table.assign(log_post_weight=log_post_weight)

	# flagging of solutions. Go through groups by primary id (IDs in first catalogue)

	table = table.assign(
		match_flag = numpy.zeros(len(table), dtype=int),
		prob_has_match = numpy.zeros(len(table)),
		prob_this_match = numpy.zeros(len(table)),
	)

	logger.log('    grouping by primary catalogue ID and flagging ...')

	def compute_group_statistics(group):
		# group
		# compute no-match probability
		values = group['log_post_weight'].values
		#values = grpvalues.values
		offset = values.max()
		bfsum = log10((10**(values - offset)).sum()) + offset
		if len(values) > 1:
			offset = values[1:].max()
			bfsum1 = log10((10**(values[1:] - offset)).sum()) + offset
		else:
			bfsum1 = 0
		assert group['ncat'].values[0] == 1, group['ncat']
		#print(group['ncat'].values[0])
		# for p_any, find the one without counterparts
		p_none = values[0]
		p_any = 1 - 10**(p_none - bfsum)
		# this avoids overflows in the no-counterpart solution, 
		# which we want to set to 0
		values[0] = bfsum1
		p_i = 10**(values - bfsum1)
		p_i[0] = 0
		
		best_val = p_i.max()
		
		# flag best & second best
		# ignore very poor solutions
		match_flag = numpy.where(best_val == p_i, 1, 
			numpy.where(p_i > prob_ratio_secondary * best_val, 2, 0))
		
		group['prob_has_match'] = p_any
		group['prob_this_match'] = p_i
		group['match_flag'] = match_flag
		return group

	table = table.groupby(table.columns[0], sort=False).apply(compute_group_statistics)
	del table['log_post_weight']
	return table

def _truncate_table(table, min_prob, logger):
	# cut away poor posteriors if requested
	if min_prob > 0:
		mask = ~(table['prob_this_match'] < min_prob)
		logger.log('    cutting away %d (below p_i minimum)' % (len(mask) - mask.sum()))
		table = table[mask]

	return table



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division

__doc__ = """Create a shifted catalogue for testing the false association rate.

Example: nway-create-shifted-catalogue.py --radius 20 --shift-ra 0 --shift-dec 60 COSMOS-XMM.fits shifted-COSMOS-XMM.fits
"""

import sys
import numpy
from numpy import log10, pi, exp, logical_and
import matplotlib.pyplot as plt
import astropy.io.fits as pyfits
import argparse
import nwaylib.progress as progress

import nwaylib.fastskymatch as match

class HelpfulParser(argparse.ArgumentParser):
	def error(self, message):
		sys.stderr.write('error: %s\n' % message)
		self.print_help()
		sys.exit(2)

parser = HelpfulParser(description=__doc__,
	epilog="""Johannes Buchner (C) 2013-2017 <johannes.buchner.acad@gmx.com>""",
	formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('--shift-dec', default=0, type=float,
	help='Shift to add in dec (arcsec)')

parser.add_argument('--shift-ra', default=0, type=float,
	help='Shift to add in ra (arcsec)')

parser.add_argument('--radius', type=float, required=True,
	help='Remove sources which are near original sources, within this radius (arcsec).')

parser.add_argument('inputfile', type=str, help="input catalogue fits file")
parser.add_argument('outputfile', help='output catalogue fits file')

# parsing arguments
args = parser.parse_args()

outfile = args.outputfile
filename = args.inputfile
radius = args.radius

print('opening', filename)
inputfitsfile = pyfits.open(filename)
header_hdu = inputfitsfile[0]
table = inputfitsfile[1].data

if args.shift_ra==0 and args.shift_dec==0:
	print('ERROR: You have to set either shift-ra or shift-dec to non-zero')
	sys.exit(1)

ra_key  = match.get_tablekeys(table, 'RA')
print('    using RA  column: %s' % ra_key)
dec_key = match.get_tablekeys(table, 'DEC')
print('    using DEC column: %s' % dec_key)

ra_orig = table[ra_key]
dec_orig = table[dec_key]

ra = ra_orig + args.shift_ra / 60. / 60
dec = dec_orig + args.shift_dec / 60. / 60

# for each of them, check that there is no collision
excluded = []

pbar = progress.bar(ndigits=6)
for i, (ra_i, dec_i) in pbar(list(enumerate(zip(ra, dec)))):
	d = match.dist((ra_i, dec_i), (ra_orig, dec_orig))
	excluded.append((d * 60 * 60 < radius).any())

excluded = numpy.array(excluded)
print('removed %d sources which collide with original positions' % (excluded.sum()))

table[ra_key] = ra
table[dec_key] = dec

newcolumns = []
for col in table.columns:
	newcolumns.append(pyfits.Column(name=col.name, format=col.format, array=col.array[~excluded]))

tbhdu = match.fits_from_columns(newcolumns)
print('writing "%s" (%d rows)' % (outfile, len(tbhdu.data)))
for k, v in inputfitsfile[1].header.items():
	if k not in tbhdu.header:
		tbhdu.header[k] = v

hdulist = pyfits.HDUList([header_hdu, tbhdu])
hdulist.writeto(outfile, **progress.kwargs_overwrite_true)






from __future__ import print_function, division

def checkupdates(current_version=None):
	"""
	Check for program updates on PyPI
	
	If you want to disable this, create a file 
	I_will_check_for_NWAY_updates_myself_thank_you 
	in your working directory.
	"""
	try:
		import os
		if os.path.exists('I_will_check_for_NWAY_updates_myself_thank_you'):
			return
		if current_version is None:
			current_version = '4.5.4'
		import json
		from StringIO import StringIO
		import urllib3
		import certifi
		http = urllib3.PoolManager(cert_reqs='CERT_REQUIRED', ca_certs=certifi.where())
		r = http.request('GET', 'https://pypi.python.org/pypi/nway/json')
		j = json.load(StringIO(r.data))
		version = j['info']['version']
		if current_version != version:
			print()
			print('A newer version of NWAY seems to be available.')
			print('Please update because it may fix bugs.')
			print()
	except Exception:
		pass
	except ImportError:
		pass
	except IOError:
		pass



"""
Probabilistic Cross-Identification of Astronomical Sources

Reference: Budavari & Szalay (2008), ApJ, 679:301-309
Authors: Johannes Buchner (C) 2013-2020
Authors: Tamas Budavari (C) 2012
"""
import numpy
import numpy.testing as test
from nwaylib.bayesdistance import log_bf, log_bf2, log_bf3, make_invcovmatrix, apply_vABv, convert_from_ellipse, log_bf_elliptical

def test_log_bf_consistent2():
	# test that log_bf function reduces to log_bf2 for 2 catalogs:
	sep = numpy.array([0., 0.1, 0.2, 0.3, 0.4, 0.5])
	for psi in sep:
		print(psi)
		print('  ', log_bf2(psi, 0.1, 0.2), )
		print('  ', log_bf([[None, psi]], [0.1, 0.2]), )
		test.assert_almost_equal(log_bf2(psi, 0.1, 0.2), log_bf([[None, psi]], [0.1, 0.2]))

def test_log_bf_consistent3():
	sep = numpy.array([0., 0.1, 0.2, 0.3, 0.4, 0.5])
	for psi in sep:
		print(psi)
		bf3 = log_bf3(psi, psi, psi, 0.1, 0.2, 0.3)
		print('  ', bf3)
		g = log_bf([[None, psi, psi], [psi, None, psi], [psi, psi, None]], [0.1, 0.2, 0.3])
		print('  ', g)
		test.assert_almost_equal(bf3, g)
	q = numpy.zeros(len(sep))
	print(log_bf(numpy.array([[numpy.nan + sep, sep, sep], [sep, numpy.nan + sep, sep], [sep, sep, numpy.nan + sep]]), 
		[0.1 + q, 0.2 + q, 0.3 + q]))

def test_ellipse_conversion():
	rng = numpy.random.RandomState(1)
	sigma_ra = rng.uniform(1, 100, size=100)
	sigma_dec = rng.uniform(1, 100, size=100)
	angles = rng.uniform(0, 180, size=100)
	# circular case:
	sigma_x, sigma_y, rho = convert_from_ellipse(sigma_ra, sigma_ra, angles)
	test.assert_almost_equal(rho, 0.)
	test.assert_almost_equal(sigma_x, sigma_ra)
	test.assert_almost_equal(sigma_y, sigma_ra)
	del sigma_x, sigma_y, rho

	# aligned case:
	sigma_x, sigma_y, rho = convert_from_ellipse(sigma_ra, sigma_dec, 0)
	test.assert_almost_equal(rho, 0.)
	test.assert_almost_equal(sigma_y, sigma_ra)
	test.assert_almost_equal(sigma_x, sigma_dec)
	del sigma_x, sigma_y, rho

	# rotated by 90 degrees
	sigma_x, sigma_y, rho = convert_from_ellipse(sigma_ra, sigma_dec, numpy.pi / 2)
	test.assert_almost_equal(rho, 0.)
	test.assert_almost_equal(sigma_x, sigma_ra)
	test.assert_almost_equal(sigma_y, sigma_dec)
	del sigma_x, sigma_y, rho

	sigma_x, sigma_y, rho_corr = convert_from_ellipse(sigma_ra, sigma_dec, angles)
	assert numpy.all(numpy.abs(rho_corr) > 1e-6), rho_corr

def test_corrmatrix_distances_circular():
	# difference in positions
	N = 10
	shift = numpy.arange(N)
	dra  = shift
	ddec = 0 * shift
	dvec = (dra, ddec)

	# precision:
	sigma_ra  = 1 * numpy.ones(N)
	sigma_dec = 1 * numpy.ones(N)
	sigma_ra2  = sigma_ra / 10
	sigma_dec2 = sigma_dec / 10
	sigma_ra_tot = (sigma_ra**-2 + sigma_ra2**-2)**-0.5
	sigma_dec_tot = (sigma_dec**-2 + sigma_dec2**-2)**-0.5

	# compute distance in units of sigma^2 by hand:
	delta1 = dra**2 / sigma_ra**2 + ddec**2 / sigma_dec**2
	delta2 = dra**2 / sigma_ra2**2 + ddec**2 / sigma_dec2**2
	delta = delta1 + delta2
	assert delta[0] == 0, delta[0]
	assert numpy.isclose(delta[1], 101), delta[1]
	test.assert_almost_equal(delta, (dra / sigma_ra_tot)**2 + (ddec / sigma_dec_tot)**2)

	# compute distance in units of sigma using bayesdistance tools:
	symsigmaI1 = make_invcovmatrix(sigma_ra, sigma_dec)
	symsigmaI2 = make_invcovmatrix(sigma_ra2, sigma_dec2)
	delta_matrix = apply_vABv(dvec, symsigmaI1, symsigmaI2)
	# make sure it is the same as the computation by hand above
	test.assert_almost_equal(delta, delta_matrix)

	# make sure the order of catalogs does not matter:
	delta_matrix2 = apply_vABv(dvec, symsigmaI2, symsigmaI1)
	test.assert_almost_equal(delta_matrix2, delta_matrix)

def test_corrmatrix_distances_elliptical():
	# difference in positions
	N = 4
	dra  = numpy.array([0., 1., 1., 1.])
	ddec = numpy.array([0., 0., 1., -1.])
	dvec = (dra, ddec)

	sigma_ra  = 1 * numpy.ones(N)
	sigma_dec = 2 * numpy.ones(N)
	sigma_ra2  = sigma_ra * 10
	sigma_dec2 = sigma_dec * 10
	#sigma_ra_tot = (sigma_ra**-2 + sigma_ra2**-2)**-0.5
	#sigma_dec_tot = (sigma_dec**-2 + sigma_dec2**-2)**-0.5
	sigma_x_corr, sigma_y_corr, rho_corr = convert_from_ellipse(sigma_ra, sigma_dec, -45 * numpy.pi / 180)
	print('circular:', sigma_ra, sigma_dec)
	print('elliptical:', sigma_x_corr, sigma_y_corr, rho_corr)

	assert numpy.all(numpy.abs(rho_corr) > 0.001), rho_corr

	symsigmaI1 = make_invcovmatrix(sigma_ra, sigma_dec)
	symsigmaI2 = make_invcovmatrix(sigma_ra2, sigma_dec2)
	delta_symm = apply_vABv(dvec, symsigmaI1, symsigmaI2)

	# use correlated errors:
	asymsigmaI1 = make_invcovmatrix(sigma_x_corr, sigma_y_corr, rho_corr)
	asymsigmaI2 = make_invcovmatrix(sigma_ra2, sigma_dec2)
	delta_corr = apply_vABv(dvec, asymsigmaI1, asymsigmaI2)
	# this must be different
	print(delta_symm, delta_corr)
	test.assert_almost_equal(delta_corr[0], delta_symm[0])
	# step along covariance should be about sqrt(2) smaller!
	assert delta_corr[2] < delta_symm[2] / 1.4, (delta_corr[2], delta_symm[2])
	# step in the orthogonal should be about sqrt(2) larger
	assert delta_corr[3] > delta_symm[3] * 1.4, (delta_corr[3], delta_symm[3])

def test_diag_mult():
	#[[None, 0]], [[None, 0]]
	dx = 1.0
	v = numpy.array([dx, 0.]), numpy.array([0., dx])
	sigma1 = 1 * numpy.ones(2)
	sigma2 = 2 * numpy.ones(2)
	prec1 = 1 / sigma1**2
	prec2 = 1 / sigma2**2
	A = make_invcovmatrix(sigma1, sigma1, 0)
	B = make_invcovmatrix(sigma2, sigma2, 0)
	print(v, A, B)
	d = apply_vABv(v, A, B)
	test.assert_almost_equal(dx * (prec1 + prec2), d)

from nwaylib.bayesdistance import vector_multiply, apply_vector_right, assert_possemdef, matrix_add

def test_ell_circ_consistent():
	f = 1
	sigma1 = 100 * numpy.ones(1)
	sigma2 = 1 * numpy.ones(1)
	A = [sigma1 * f, sigma1 * f, 0]
	B = [sigma2 * f, sigma2 * f, 0]

	# identical circular errors, co-centered
	bf_ell = log_bf_elliptical([[None, 0]], [[None, 0]], [B, B])
	bf_circ = log_bf([[None, 0]], [sigma2,sigma2])
	test.assert_almost_equal(bf_ell, bf_circ, decimal=5)
	print("---")

	# identical circular errors, co-centered 2
	bf_ell = log_bf_elliptical([[None, 0]], [[None, 0]], [A, A])
	bf_circ = log_bf([[None, 0]], [sigma1,sigma1])
	test.assert_almost_equal(bf_ell, bf_circ, decimal=5)
	print("---")

	step = 1.0 * numpy.ones(1)
	dstep = step # * 2**0.5

	print("---")
	# equally sized pos errors, off centered
	bf_ell = log_bf_elliptical([[None, dstep]], [[None, 0]], [A, A])
	bf_circ = log_bf([[None, step]], [sigma1, sigma1])
	test.assert_almost_equal(bf_ell, bf_circ)

	print("---")
	# equally sized pos errors, off centered
	bf_ell = log_bf_elliptical([[None, dstep]], [[None, 0]], [B, B])
	bf_circ = log_bf([[None, step]], [sigma2, sigma2])
	test.assert_almost_equal(bf_ell, bf_circ)

	print("---")
	# different pos errors per catalog:
	bf_ell = log_bf_elliptical([[None, dstep]], [[None, 0]], [A, B])
	print(bf_ell)
	bf_circ = log_bf([[None, step]], [sigma1, sigma2])
	test.assert_almost_equal(bf_ell, bf_circ)

	print("---")
	bf_ell = log_bf_elliptical([[None, step]], [[None, 0]], [convert_from_ellipse(0.1,0.1,0), convert_from_ellipse(0.2,0.2,0)])
	bf_circ = log_bf([[None, dstep]], [0.1,0.2])
	test.assert_almost_equal(bf_ell, bf_circ, decimal=5)

	bf_ell2 = log_bf_elliptical([[None, 2*step]], [[None, 0]], [convert_from_ellipse(0.1,0.1,0), convert_from_ellipse(0.2,0.2,0)])
	bf_circ2 = log_bf([[None, 2*dstep]], [0.1,0.2])
	test.assert_almost_equal(bf_ell2, bf_circ2, decimal=5)
	
	print(bf_ell2, bf_ell)

	bf_ell = log_bf_elliptical([[None, 1.]], [[None, 1.]], [convert_from_ellipse(0.1,0.1,0), convert_from_ellipse(0.2,0.2,0)])
	bf_circ = log_bf([[None, 2**0.5]], [0.1,0.2])
	test.assert_almost_equal(bf_ell, bf_circ)


def test_ell_semdef_pos():
	N = 400
	sigma_ra = 1.0
	sigma_dec = numpy.random.uniform(size=N)
	sigma_x_corr, sigma_y_corr, rho_corr = convert_from_ellipse(sigma_ra, sigma_dec, numpy.random.uniform(0, 2 * numpy.pi))
	symsigmaI1 = make_invcovmatrix(sigma_x_corr, sigma_y_corr, rho_corr)
	assert_possemdef(symsigmaI1)
	
	sigma2_ra = numpy.random.uniform(size=N)
	sigma2_dec = numpy.random.uniform(size=N)
	sigma2_x_corr, sigma2_y_corr, rho2_corr = convert_from_ellipse(sigma2_ra, sigma2_dec, numpy.random.uniform(0, 2 * numpy.pi))
	symsigmaI2 = make_invcovmatrix(sigma2_x_corr, sigma2_y_corr, rho2_corr)
	assert_possemdef(symsigmaI2)
	assert_possemdef(matrix_add(symsigmaI1, symsigmaI2))

	v = numpy.random.normal(0, 1, size=(2, N))
	q = apply_vABv(v, symsigmaI1, symsigmaI2)
	q2 = vector_multiply(v, apply_vector_right(matrix_add(symsigmaI1, symsigmaI2), v))
	print(q, (q>=0).all())
	print(q2, (q2>=0).all())
	assert (q>=0).all()
	assert (q2>=0).all()



import numpy
from numpy import pi, sin, exp, cos, log, log10

fout = open('randomcatO.csv', 'wb')
fout.write(b'ID,RA,DEC\n')

rarange  = numpy.arange(150.0, 150.1, 1./60./60.)
decrange = numpy.arange(2.0, 2.1,     1./60./60.)

print('area: %.3f deg2' % ((rarange[-1] - rarange[0]) * (decrange[-1] - decrange[0])))

for i, ra in enumerate(rarange):
	ids = 1 + i*len(decrange) + numpy.arange(len(decrange))
	numpy.savetxt(fout, numpy.transpose([ids,numpy.random.normal(decrange, 0.01/60/60)]), fmt='%%d,%.6f,%%.6f' % ra)

fout.close()

fout = open('randomcatX.csv', 'w')
fout2 = open('randomcatR.csv', 'w')
fout.write('ID,RA,DEC,pos_err,a,b,phi\n')
fout2.write('ID,RA,DEC,pos_err,a,b,phi\n')

id = 1
ractr  = rarange.mean()
decctr = decrange.mean()

lastr = 0.
for i in range(10):
	r = (1. + i**2) / 60. / 60.
	
	if i % 2 == 0:
		a  = (r - lastr) / 10.
		b  = (r - lastr) / 10.
	else:
		a  = (r - lastr) / 10. * 2
		b  = (r - lastr) / 10. / 2
	symerror = (a**2 + b**2)**0.5
	
	for j, angle in enumerate(numpy.arange(0, 2*pi, pi/6)):
		
		ra  = ractr + r * cos(angle)
		dec = decctr + r * sin(angle)
		rotangle = angle if (i+1) % 4 == 0 else angle + pi/2
		
		fout.write("%d,%.6f,%.6f,%.6f,%.6f,%.6f,%.6f\n" % (id, ra, dec, symerror*60*60,  a*60*60, b*60*60, rotangle/pi*180))
		fout2.write("%d,%.6f,%.6f,%.6f,%.6f,%.6f,%.6f\n" % (id, numpy.random.normal(ra, symerror), numpy.random.normal(dec, symerror), symerror*60*60,  a*60*60, b*60*60, rotangle/pi*180))
		id += 1
	lastr = r
fout.close()
fout2.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division

__doc__ = """Characterise the false association rate and efficiency of a match
with a offset (fake) match and a real match.

Example: nway-calibrate-cutoff.py example2.fits example2-shifted-match.fits
"""

import sys
import numpy
from numpy import log10, pi, exp, logical_and
import matplotlib.pyplot as plt
import astropy.io.fits as pyfits
import argparse

import nwaylib.fastskymatch as match

class HelpfulParser(argparse.ArgumentParser):
	def error(self, message):
		sys.stderr.write('error: %s\n' % message)
		self.print_help()
		sys.exit(2)

parser = HelpfulParser(description=__doc__,
	epilog="""Johannes Buchner (C) 2013-2017 <johannes.buchner.acad@gmx.com>""",
	formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('realfile', help="match output using real catalogue")
parser.add_argument('fakefile', help='match output using fake catalogue')

# parsing arguments
args = parser.parse_args()

f = pyfits.open(args.realfile)
data = f[1].data
mask = data['ncat'] == 1
p_any0 = data['p_any'][mask]
mask = data['match_flag'] == 1
p_any = data['p_any'][mask]
p_i = data['p_i'][mask]

f = pyfits.open(args.fakefile)
data = f[1].data
mask = data['ncat'] == 1
p_any0_offset = data['p_any'][mask]
mask = data['match_flag'] == 1
p_any_offset = data['p_any'][mask]
p_i_offset = data['p_i'][mask]

cutoffs = numpy.linspace(0, 1, 101)

efficiency = [(p_any0 > cutoff).mean() for cutoff in cutoffs]
error_rate = [(p_any0_offset > cutoff).mean() for cutoff in cutoffs]

plt.figure(figsize=(5,5))
plt.plot(cutoffs, efficiency, '-', color='k', label='selection efficiency', lw=2)
plt.plot(cutoffs, error_rate, '--', color='r', label='false selection rate')

plt.ylabel('fraction(>p_any)')
plt.xlabel('p_any')
plt.legend(loc='lower left', prop=dict(size=10))
plt.savefig(args.realfile + '_p_any_cutoffquality.pdf', bbox_inches='tight')
plt.savefig(args.realfile + '_p_any_cutoffquality.png', bbox_inches='tight')
plt.close()
print('created plot "%s_p_any_cutoffquality.pdf"' % args.realfile)

plt.figure(figsize=(5,5))
plt.plot(p_any, p_i, '. ', color='r', label='real')
plt.plot(p_any_offset, p_i_offset, '. ', color='gray', label='offset')

plt.ylabel('p_i')
plt.xlabel('p_any')
plt.legend(loc='lower left', prop=dict(size=10))
plt.savefig(args.realfile + '_p_any_p_i.pdf', bbox_inches='tight')
plt.savefig(args.realfile + '_p_any_p_i.png', bbox_inches='tight')
plt.close()
print('created plot "%s_p_any_p_i.pdf"' % args.realfile)

for rate in [0.01,0.03,0.05,0.1]:
	print()
	# find where error-rate is < rate
	mask = numpy.array(error_rate) < rate
	if not mask.any():
		print('A false detection rate of <%d%% is not possible.' % (rate*100))
	else:
		i = numpy.min(numpy.where(mask)[0])
		#print(i)
		print('For a false detection rate of <%d%%' % (rate*100))
		print('--> use only counterparts with p_any>%.2f (%.2f%% of matches)' % (cutoffs[i], efficiency[i] * 100))



"""
Computes adaptively binned histograms of two distributions.

Provides smooth biasing functions and plots them.

This is not specific for magnitudes, it can be any provided property.
"""

import scipy, scipy.interpolate, scipy.signal, scipy.integrate
import numpy
from numpy import log10, pi, exp, logical_and
import matplotlib.pyplot as plt

"""
ratio between the two histograms. Minor offsets to avoid zero and inf.
"""
def ratio(hist_sel, hist_all):
	with numpy.errstate(divide='ignore', invalid='ignore'):
		return numpy.where(hist_all == 0, 100, hist_sel / hist_all)

"""
fraction of selected
"""
def fraction(bin_mag, hist_sel, hist_all):
	# need a re-normalisation so that unknown values can be set to prob=1
	m = hist_all > 0
	hist_sel_m = hist_sel[m]
	hist_all_m = hist_all[m]
	ratio = hist_sel_m / hist_all_m
	delta = (bin_mag[1:] - bin_mag[:-1])[m]
	avg = (ratio * hist_all_m * delta).sum() / (delta * hist_all_m).sum()
	assert avg != 0
	
	r = numpy.ones(len(hist_all))
	r[m] = hist_sel_m / hist_all_m / avg
	assert numpy.isfinite(r).all(), r[~numpy.isfinite(r)]
	assert (r > 0).all(), r[~(r > 0)]
	return r

"""
Plotting
"""
def plot_fit(bin_mag, hist_sel, hist_all, func, name):
	mags = numpy.linspace(bin_mag.min(), bin_mag.max(), 400)
	plt.figure()
	plt.subplot(2, 1, 1)
	hist_n = ratio(hist_sel, hist_all)
	plt.plot(bin_mag[:-1], hist_all, '-', 
		drawstyle='steps-post', label='all')
	plt.plot(bin_mag[:-1], hist_sel, '-', 
		drawstyle='steps-post', label='selected')
	plt.legend(loc='best')
	plt.ylabel('normalized weight')
	plt.xlim(mags.min(), mags.max())
	plt.subplot(2, 1, 2)
	plt.plot(bin_mag[:-1], hist_n, '-',
		drawstyle='steps-post', label='ratio')
	plt.plot(mags, func(mags), '-', label='fit')
	plt.legend(loc='best')
	plt.ylabel('normalized weight')
	plt.xlabel(name)
	plt.xlim(mags.min(), mags.max())
	plt.yscale('log')
	plt.savefig(name.replace(':', '_') + '_fit.pdf', bbox_inches='tight')
	plt.close()

"""
creates the biasing functions
"""
def fitfunc_histogram(bin_mag, hist_sel, hist_all):
	bin_n = ratio(hist_sel, hist_all)
	# w = scipy.signal.gaussian(5, 1)
	# w /= w.sum()
	# bin_n_smooth = scipy.signal.convolve(bin_n, w, mode='same')
	# no smoothing
	bin_n_smooth = bin_n
	interpfunc = scipy.interpolate.interp1d(bin_mag,
		list(bin_n_smooth) + [bin_n_smooth[-1]],
		bounds_error=False, kind='zero')
	return interpfunc

"""
creates the histograms for the two columns in an adaptive way (based on mag_sel)
with the same binning.
"""
def adaptive_histograms(mag_all, mag_sel, weights=None):
	if weights is None:
		weights = numpy.ones(len(mag_sel))
	assert len(weights) == len(mag_sel), (len(weights), len(mag_sel))
	mag_sel_idx = numpy.argsort(mag_sel)
	mag_sel_sorted = mag_sel[mag_sel_idx]
	weight_axis = numpy.cumsum(weights[mag_sel_idx]) / numpy.sum(weights)
	weight_axis[0] = 0
	weight_axis[-1] = 1
	#weight_axis = numpy.linspace(0, 1, len(mag_sel_sorted))
	
	func_sel = scipy.interpolate.interp1d(weight_axis, mag_sel_sorted)
	# choose bin borders based on cumulative distribution, using 15 points
	x = numpy.unique(func_sel(numpy.linspace(0, 1, 15)))
	lo, hi = numpy.nanmin(mag_all), numpy.nanmax(mag_all)
	# extend to make sure ratios are well-defined
	if x[-1] < hi:
		x = numpy.asarray(list(x) + [hi+1])
	if x[0] > lo:
		x = numpy.asarray([lo-1] + list(x))
	# linear histogram (for no adaptiveness):
	##x = numpy.linspace(mag_all.min(), mag_all.max(), 20)
	hist_sel, bins = numpy.histogram(mag_sel, bins=x,    density=True, weights=weights)
	hist_all, bins = numpy.histogram(mag_all, bins=bins, density=True)
	return bins, hist_sel, hist_all





"""
Functions for finding pairs within a certain search distance "err".

Very fast method based on hashing.
"""
from __future__ import print_function, division
import numpy
import itertools
from collections import defaultdict
import os
import astropy.io.fits as pyfits
from astropy.coordinates import SkyCoord, SkyOffsetFrame
import astropy.units as u
from numpy import sin, cos, arctan2, hypot, arccos, arcsin, pi, exp, log
import healpy
import tqdm
import joblib
cachedir = 'cache'
if not os.path.isdir(cachedir): os.mkdir(cachedir)
mem = joblib.Memory(cachedir, verbose=False)

def dist(apos, bpos):
	"""
	Angular separation (in degrees) between two points on a sphere.
	http://en.wikipedia.org/wiki/Great-circle_distance
	"""
	(a_ra, a_dec), (b_ra, b_dec) = apos, bpos
	lon1 = a_ra / 180 * pi
	lat1 = a_dec / 180 * pi
	lon2 = b_ra / 180 * pi
	lat2 = b_dec / 180 * pi
	sdlon = sin(lon2 - lon1)
	cdlon = cos(lon2 - lon1)
	slat1 = sin(lat1)
	slat2 = sin(lat2)
	clat1 = cos(lat1)
	clat2 = cos(lat2)

	num1 = clat2 * sdlon
	num2 = clat1 * slat2 - slat1 * clat2 * cdlon
	denominator = slat1 * slat2 + clat1 * clat2 * cdlon

	return arctan2(hypot(num1, num2), denominator) * 180 / pi

def dist3d(apos, bpos):
	"""
	Angular separation in ra & dec between two points on a sphere.
	"""
	(a_ra, a_dec), (b_ra, b_dec) = apos, bpos
	a_ra  = numpy.where(a_ra  == -99, numpy.nan, a_ra)
	a_dec = numpy.where(a_dec == -99, numpy.nan, a_dec)
	b_ra  = numpy.where(b_ra  == -99, numpy.nan, b_ra)
	b_dec = numpy.where(b_dec == -99, numpy.nan, b_dec)
	with numpy.errstate(divide='ignore', invalid='ignore'):
		a = SkyCoord(a_ra, a_dec, frame="icrs", unit="deg")
		b = SkyCoord(b_ra, b_dec, frame="icrs", unit="deg")
	
		localframe = SkyOffsetFrame(origin=a)
		na = a.transform_to(localframe)
		nb = b.transform_to(localframe)
		dra  = na.lon - nb.lon
		ddec = na.lat - nb.lat

		separation = a.separation(b).to(u.degree).value
	
		dra  = dra.to(u.degree).value
		ddec = ddec.to(u.degree).value
	
		return separation, dra, ddec

def get_tablekeys(table, name, tablename=''):
	keys = sorted(table.dtype.names, key=lambda k: 0 if k.upper() == name else 1 if k.upper().startswith(name) else 2)
	assert len(keys) > 0 and name in keys[0].upper(), 'ERROR: No "%s" column found in input catalogue "%s". Only have: %s' % (name, tablename, ', '.join(table.dtype.names))
	return keys[0]


def get_healpix_resolution_degrees(nside):
	resol = healpy.pixelfunc.nside2resol(nside) / pi * 180
	# according to monte carlo simulations, distances up to this factor
	# are completely contained within the pixel and its neighbors
	resfactor = 0.7
	return resfactor * resol

@mem.cache(ignore=['logger'])
def crossproduct(radectables, err, logger, pairwise_errs=[]):
	# check if away from the poles and RA=0
	use_flat_bins = True
	for ra, dec in radectables:
		if not(err < 1 and (ra > 10*err).all() and (ra < 360-10*err).all() and (numpy.abs(dec) < 45).all()):
			use_flat_bins = False
			break
	
	if use_flat_bins:
		logger.log('matching: using fast flat-sky approximation for this match')
	else:
		# choose appropriate nside for err (in deg)
		nside = 1
		for nside_next in range(30): 
			# largest distance still contained within pixels
			dist_neighbors_complete = get_healpix_resolution_degrees(2**nside_next)
			# we are looking for a pixel size which ensures bigger distances than the error radius
			# but we want the smallest pixels possible, to reduce the cartesian product
			if dist_neighbors_complete < err:
				# too small, do not accept
				# sources within err will be outside the neighbor pixels
				break
			nside = 2**nside_next
		resol = get_healpix_resolution_degrees(nside) * 60 * 60
		logger.log('matching: healpix hashing on pixel resolution ~ %f arcsec (nside=%d)' % (resol, nside))
	
	buckets = defaultdict(lambda : [[] for _ in range(len(radectables))])
	primary_cat_keys = None
	
	pbar = tqdm.tqdm(total=sum([len(t[0]) for t in radectables]))
	for ti, (ra_table, dec_table) in enumerate(radectables):
		if use_flat_bins:
			for ei, (ra, dec) in enumerate(zip(ra_table, dec_table)):
				i, j = int(ra / err), int(dec / err)
				
				# put in bucket, and neighbors
				for jj, ii in (j,i), (j,i+1), (j+1,i), (j+1, i+1):
					k = (ii, jj)
					# only primary catalogue is allowed to define new buckets
					if ti == 0 or k in buckets:
						buckets[k][ti].append(ei)
				pbar.update()
		else:
			# get healpixels
			ra, dec = ra_table, dec_table
			phi = ra / 180 * pi
			theta = dec / 180 * pi + pi/2.
			i = healpy.pixelfunc.ang2pix(nside, phi=phi, theta=theta, nest=True)
			j = healpy.pixelfunc.get_all_neighbours(nside, phi=phi, theta=theta, nest=True)
			# only consider four neighbours in one direction (N)
			# does not work, sometimes A is south of B, but B is east of A
			# so need to consider all neighbors, and deduplicate later
			neighbors = numpy.hstack((i.reshape((-1,1)), j.transpose()))
			
			# put in bucket, and neighbors
			if ti == 0:
				# only primary catalogue is allowed to define new buckets
				for ei, keys in enumerate(neighbors):
					for k in keys:
						buckets[k][ti].append(ei)
					pbar.update()
			else:
				for ei, keys in enumerate(neighbors):
					for k in keys:
						if k in primary_cat_keys:
							buckets[k][ti].append(ei)
					pbar.update()
			if ti == 0:
				primary_cat_keys = set(buckets.keys())
				
	pbar.close()
	
	# add no-counterpart options
	results = set()
	# now combine within buckets
	logger.log('matching: collecting from %d buckets, creating cartesian products ...' % len(buckets))
	#print('matching: %6d matches expected after hashing' % numpy.sum([
	#	len(lists[0]) * numpy.product([len(li) + 1 for li in lists[1:]]) 
	#		for lists in buckets.values()]))
	
	#pbar = logger.progress(ndigits=5, maxval=len(buckets)).start()
	pbar = tqdm.tqdm(total=len(buckets))
	while buckets:
		k, lists = buckets.popitem()
		pbar.update()
		# add for secondary catalogues the option of missing source
		for l in lists[1:]:
			l.insert(0, -1)
		# create the cartesian product
		local_results = itertools.product(*[sorted(l) for l in lists])
		
		# if pairwise filtering is requested, use it to trim down solutions
		if pairwise_errs:
			local_results = numpy.array(list(local_results))
			#nstart = len(local_results)
			for tablei, tablej, errij in pairwise_errs:
				indicesi = local_results[:,tablei]
				indicesj = local_results[:,tablej]
				# first find entries that actually have both entries
				mask_both = numpy.logical_and(indicesi >= 0, indicesj >= 0)
				#if not mask_both.any():
				#	continue
				# get the RA/Dec
				rai, deci = radectables[tablei]
				raj, decj = radectables[tablej]
				rai, deci = rai[indicesi[mask_both]], deci[indicesi[mask_both]]
				raj, decj = raj[indicesj[mask_both]], decj[indicesj[mask_both]]
				# compute distances
				mask_good = dist((rai, deci), (raj, decj)) < errij * 60 * 60
				# select the ones where one is missing, or those within errij
				mask_good2 = ~mask_both
				mask_good2[mask_both][mask_good] = True
				#print(mask_good2.sum(), mask_both.shape, mask_both.sum(), mask_good.sum())
				local_results = local_results[mask_good2,:]
			
			#print("compression:%d/%d" % (len(local_results), nstart))
			results.update([tuple(l) for l in local_results])
		else:
			results.update(local_results)
		del local_results
	pbar.close()

	n = len(results)
	logger.log('matching: %6d unique matches from cartesian product. sorting ...' % n)
	# now make results unique by sorting
	results = numpy.array(sorted(results))
	return results

# use preferred newer astropy command if available
if hasattr(pyfits.BinTableHDU, 'from_columns'):
	fits_from_columns = pyfits.BinTableHDU.from_columns
else:
	fits_from_columns = pyfits.new_table

def match_multiple(tables, table_names, err, fits_formats, logger, circular=True, pairwise_errs=[]):
	"""
	computes the cartesian product of all possible matches,
	limited to a maximum distance of err (in degrees).
	
	tables: input FITS table
	table_names: names of the tables
	fits_formats: FITS data type of the columns of each table
	
	returns 
	results: cartesian product of all possible matches (smaller than err)
	cat_columns: table with separation distances in arcsec
	header: which columns were used in each table for RA/DEC
	
	"""
	logger.log('')
	logger.log('matching with %f arcsec radius' % (err * 60 * 60))
	logger.log('matching: %6d naive possibilities' % numpy.product([len(t) for t in tables]))

	logger.log('matching: hashing')

	ra_keys = [get_tablekeys(table, 'RA', tablename=tablename) for table, tablename in zip(tables, table_names)]
	logger.log('    using RA  columns: %s' % ', '.join(ra_keys))
	dec_keys = [get_tablekeys(table, 'DEC', tablename=tablename) for table, tablename in zip(tables, table_names)]
	logger.log('    using DEC columns: %s' % ', '.join(dec_keys))

	ratables = [(t[ra_key], t[dec_key]) for t, ra_key, dec_key in zip(tables, ra_keys, dec_keys)]
	resultstable = crossproduct(ratables, err, logger=logger, pairwise_errs=pairwise_errs)
	results = resultstable.view(dtype=[(table_name, resultstable.dtype) for table_name in table_names]).reshape((-1,))

	keys = []
	for table_name, table in zip(table_names, tables):
		keys += ["%s_%s" % (table_name, n) for n in table.dtype.names]
	
	logger.log('merging in %d columns from input catalogues ...' % sum([1 + len(table.dtype.names) for table in tables]))
	cat_columns = []
	pbar = tqdm.tqdm(total=sum([1 + len(table.dtype.names) for table in tables]))
	for table, table_name, fits_format in zip(tables, table_names, fits_formats):
		tbl = table[results[table_name]]
		# set missing to nan
		mask_missing = results[table_name] == -1
		
		pbar.update()
		for n, format in zip(table.dtype.names, fits_format):
			k = "%s_%s" % (table_name, n)
			keys.append(k)
			col = tbl[n]
			#print('   setting "%s" to -99 (%d affected; column format "%s")' % (k, mask_missing.sum(), format))
			try:
				col[mask_missing] = -99
			except Exception as e:
				logger.log('   setting "%s" to -99 failed (%d affected; column format "%s"): %s' % (k, mask_missing.sum(), format, e))
			
			fitscol = pyfits.Column(name=k, format=format, array=col)
			cat_columns.append(fitscol)
			pbar.update()
	pbar.close()
	
	tbhdu = fits_from_columns(pyfits.ColDefs(cat_columns))
	header = dict(
		COLS_RA = ' '.join(["%s_%s" % (ti, ra_key) for ti, ra_key in zip(table_names, ra_keys)]),
		COLS_DEC = ' '.join(["%s_%s" % (ti, dec_key) for ti, dec_key in zip(table_names, dec_keys)])
	)
	
	logger.log('    adding angular separation columns')
	max_separation = numpy.zeros(len(results))
	for i in range(len(tables)):
		a_ra  = tbhdu.data["%s_%s" % (table_names[i], ra_keys[i])]
		a_dec = tbhdu.data["%s_%s" % (table_names[i], dec_keys[i])]
		for j in range(i):
			k = "Separation_%s_%s" % (table_names[i], table_names[j])
			k1 = k + "_ra"
			k2 = k + "_dec"
			if circular:
				keys += [k, k1, k2]
			else:
				keys += [k]
			
			b_ra  = tbhdu.data["%s_%s" % (table_names[j], ra_keys[j])]
			b_dec = tbhdu.data["%s_%s" % (table_names[j], dec_keys[j])]
			
			if circular:
				col = dist((a_ra, a_dec), (b_ra, b_dec))
			else:
				col, col_ra, col_dec = dist3d((a_ra, a_dec), (b_ra, b_dec))
			
			valid_input = numpy.logical_and(a_ra != -99, b_ra != -99)
			assert not numpy.isnan(col[valid_input]).any(), ['%d distances are nan' % numpy.isnan(col[valid_input]).sum(), 
				a_ra[numpy.isnan(col)], a_dec[numpy.isnan(col)], 
				b_ra[numpy.isnan(col)], b_dec[numpy.isnan(col)]]
			
			col[a_ra == -99] = numpy.nan
			col[b_ra == -99] = numpy.nan
			if not circular:
				col_ra[a_ra == -99] = numpy.nan
				col_ra[b_ra == -99] = numpy.nan
				col_dec[a_ra == -99] = numpy.nan
				col_dec[b_ra == -99] = numpy.nan
			max_separation = numpy.nanmax([col * 60 * 60, max_separation], axis=0)
			# store distance in arcsec 
			cat_columns.append(pyfits.Column(name=k, format='E', array=col * 60 * 60))
			if not circular:
				cat_columns.append(pyfits.Column(name=k1, format='E', array=col_ra * 60 * 60))
				cat_columns.append(pyfits.Column(name=k2, format='E', array=col_dec * 60 * 60))
	
	cat_columns.append(pyfits.Column(name="Separation_max", format='E', array=max_separation))
	cat_columns.append(pyfits.Column(name="ncat", format='I', array=(resultstable > -1).sum(axis=1)))
	keys.append("Separation_max")
	mask = max_separation < err * 60 * 60
	for c in cat_columns:
		c.array = c.array[mask]
	
	logger.log('matching: %6d matches after filtering by search radius' % mask.sum())
	logger.log('')
	return results[mask], cat_columns, header

def wraptable2fits(cat_columns, extname):
	tbhdu = fits_from_columns(pyfits.ColDefs(cat_columns))
	hdu = pyfits.PrimaryHDU()
	import datetime, time
	now = datetime.datetime.fromtimestamp(time.time())
	nowstr = now.isoformat()
	nowstr = nowstr[:nowstr.rfind('.')]
	hdu.header['DATE'] = nowstr
	hdu.header['ANALYSIS'] = 'NWAY matching'
	tbhdu.header['EXTNAME'] = extname
	hdulist = pyfits.HDUList([hdu, tbhdu])
	return hdulist

def array2fits(table, extname):
	cat_columns = pyfits.ColDefs([pyfits.Column(name=n, format='E',array=table[n]) 
		for n in table.dtype.names])
	return wraptable2fits(cat_columns, extname)



import os
try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

long_description = ""
with open('README.rst') as f:
	long_description = f.read()

setup(
	name='nway',
	version='4.5.4',
	author='Johannes Buchner',
	author_email='johannes.buchner.acad@gmx.com',
	packages=['nwaylib'],
	scripts=['nway.py', 'nway-write-header.py', 'nway-explain.py', 'nway-create-fake-catalogue.py', 'nway-create-shifted-catalogue.py', 'nway-calibrate-cutoff.py'],
	url='http://pypi.python.org/pypi/nway/',
	license='AGPLv3 (see LICENSE file)',
	description='Probabilistic Cross-Identification of Astronomical Sources',
	long_description=long_description,
	install_requires=[
		"scipy",
		"numpy",
		"astropy",
		"tqdm",
		"matplotlib",
		"joblib",
		"healpy",
		"pandas",
	],
	setup_requires=['pytest-runner'],
	tests_require=['pytest'],
)
